#include "impl.h"

/*
 * Copyright 2010 Sven Vermeulen.
 * Subject to the GNU Public License, version 3.
 */
 
/***********************************************************************************************
 * Helper functions for the sqlite3 database manipulations
 ***********************************************************************************************/

sqlite3 * get_local_db(struct workstate * ws, char part, int len) {
	if (part == 'a') {
		return ws->localdb[len];
	} else if (part == 'h') {
		return ws->localdb[64+len];
	} else if (part == 'o') {
		return ws->localdb[2*64+len];
	};

	return NULL;
};

int get_db_intresult(void * cbobj, int argc, char **argv, char **azColName) {
	int i = 0;
	struct workstate * ws = (struct workstate *) cbobj;

	for(i=0;i<argc;i++) {
		// Found result. Storing in workstate variable
		int ret = atoi(argv[i]);
		ws->rc=ret;
		return 0;
	};

	return 0;
};

int get_cpe_dataresult(void * cbobj, int argc, char **argv, char **azColName) {
	int i = 0;
	struct workstate * ws = (struct workstate *) cbobj;

	for(i = 0; i < argc; i++) {
		if (strcmp(azColName[i], "cpepart") == 0) {
			ws->cpebuffer.part = argv[i][0];
			continue;
		}
		if (strcmp(azColName[i], "cpevendor") == 0) {
			strncpy(ws->cpebuffer.vendor, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeproduct") == 0) {
			strncpy(ws->cpebuffer.product, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeversion") == 0) {
			strncpy(ws->cpebuffer.version, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeupdate") == 0) {
			strncpy(ws->cpebuffer.update, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpeedition") == 0) {
			strncpy(ws->cpebuffer.edition, argv[i], 64);
			continue;
		}
		if (strcmp(azColName[i], "cpelanguage") == 0) {
			strncpy(ws->cpebuffer.language, argv[i], 64);
			continue;
		};
	};

	ws->rc = 0;
	return 0;
};

int get_cpe_data(sqlite3 * db, char * stmt, struct workstate * ws) {
	int rc = 0;
	char * errmsg;

	ws->rc = 0;
	rc = sqlite3_exec(db, stmt, get_cpe_dataresult, ws, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to execute CPE-returning statement against database: %s\n", errmsg);
		sqlite3_free(errmsg);
	};

	return ws->rc;
};

int get_int_value(sqlite3 * db, char * stmt, struct workstate * ws) {
	int rc = 0;
	char * errmsg;

	ws->rc = 0;
	rc = sqlite3_exec(db, stmt, get_db_intresult, ws, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to execute INT-returning statement against database: %s\n", errmsg);
		sqlite3_free(errmsg);
	};

	return ws->rc;
};

void run_statement(struct workstate * ws, sqlite3 * db, char * stmt) {
	int rc = 0;
	char * errmsg;

	rc = sqlite3_exec(db, stmt, NULL, NULL, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to run statement: %s\n", errmsg);
		sqlite3_free(errmsg);
	};
};

void run_statement_alldb(struct workstate * ws, char * stmt) {
	int i = 0;

	for (i = 0; i < 3*64+1; i++) {
		run_statement(ws, ws->localdb[i], stmt);
	};
};

/**
 * Clear the version database.
 */
void dbimpl_clear_versiondatabase(struct workstate * ws) {
	char stmt[1024];

	sprintf(stmt, "delete from tb_binmatch;");

	run_statement(ws, ws->localdb[0], stmt);
};

/**
 * Clear the version gathering data
 */
void dbimpl_clear_versiondata(struct workstate * ws) {
	char stmt[1024];

	sprintf(stmt, "delete from tb_versionmatch;");

	run_statement(ws, ws->matchdb, stmt);

	sprintf(stmt, "delete from tb_cpe;");

	run_statement(ws, ws->matchdb, stmt);
};

/**
 * Load the databases into the workstate
 *
 * The databases that are needed for the SQLite3 implementation are loaded in
 * the workstate variable. 
 */
int dbimpl_load_databases(struct workstate * ws) {
	const config_setting_t * localdb;
	const config_setting_t * globaldb;
	int rc = 0;
	int i = 0;
	char buffer[512];
	char buffer2[512];

	/*
	 * Load the set of databases
	 */
	localdb = config_lookup(ws->cfg, "localdb");
	if (localdb == NULL) {
		fprintf(stderr, "Configuration file does not contain localdb directive.\n");
		return 1;
	};
	rc = strlen(config_setting_get_string(localdb));
	if (rc > 500) {
		fprintf(stderr, "Configuration files 'localdb' directive cannot exceed 500 characters\n");
		return 1;
	};
	strncpy(buffer, config_setting_get_string(localdb), rc);
	buffer[rc] = 0x00;
	strcat(buffer, "/");

	sprintf(buffer2, "%smain.db", buffer);
	rc = sqlite3_open(buffer2, &(ws->localdb[0]));
	if (rc) {
		fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[0]));
		sqlite3_close(ws->localdb[0]);
		return rc;
	};
	run_statement(ws, ws->localdb[0], "PRAGMA cache_size=10000;");
	run_statement(ws, ws->localdb[0], "PRAGMA synchronous=OFF;");

	// a-part
	for (i = 1; i <= 64; i++) {
		sprintf(buffer2, "%sa%d.db", buffer, i);
		rc = sqlite3_open(buffer2, &(ws->localdb[i]));
		if (rc) {
			fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[i]));
			sqlite3_close(ws->localdb[i]);
			return rc;
		};
		run_statement(ws, ws->localdb[i], "PRAGMA cache_size=10000;");
		run_statement(ws, ws->localdb[i], "PRAGMA synchronous=OFF;");
	};
	// h-part
	for (i = 1; i <= 64; i++) {
		sprintf(buffer2, "%sh%d.db", buffer, i);
		rc = sqlite3_open(buffer2, &(ws->localdb[64+i]));
		if (rc) {
			fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[64+i]));
			sqlite3_close(ws->localdb[64+i]);
			return rc;
		};
		run_statement(ws, ws->localdb[64+i], "PRAGMA cache_size=10000;");
		run_statement(ws, ws->localdb[64+i], "PRAGMA synchronous=OFF;");
	};
	// o-part
	for (i = 1; i <= 64; i++) {
		sprintf(buffer2, "%so%d.db", buffer, i);
		rc = sqlite3_open(buffer2, &(ws->localdb[128+i]));
		if (rc) {
			fprintf(stderr, "Can't open database %s: %s\n", buffer2, sqlite3_errmsg(ws->localdb[128+i]));
			sqlite3_close(ws->localdb[128+i]);
			return rc;
		};
		run_statement(ws, ws->localdb[128+i], "PRAGMA cache_size=10000;");
		run_statement(ws, ws->localdb[128+i], "PRAGMA synchronous=OFF;");
	};
	
	globaldb = config_lookup(ws->cfg, "globaldb");
	if (globaldb == NULL) {
		fprintf(stderr, "Configuration file does not contain globaldb directive.\n");
		return 1;
	};
	rc = strlen(config_setting_get_string(globaldb));
	if (rc > 500) {
		fprintf(stderr, "Configuration files 'globaldb' directive cannot exceed 500 characters\n");
		return 1;
	};
	zero_string(buffer, 512);
	strncpy(buffer, config_setting_get_string(globaldb), rc);
	rc = sqlite3_open(buffer, &(ws->matchdb));
	if (rc) {
		fprintf(stderr, "Can't open database %s: %s\n", buffer, sqlite3_errmsg(ws->matchdb));
		sqlite3_close(ws->matchdb);
		return rc;
	};
	run_statement(ws, ws->matchdb, "PRAGMA cache_size=10000;");
	run_statement(ws, ws->matchdb, "PRAGMA synchronous=OFF;");

	return rc;
};

void add_to_database(struct workstate * ws, struct cpe_data cpe) {
	int rc = 0;
	char stmt[1024];

	// Full match

	sprintf(stmt, "select count(*) from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);

	rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);

	if (rc == 0) {
		sprintf(stmt, "insert into tb_cpe_%c_%zu (cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage) values (\"%c\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\");", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		run_statement(ws, get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt);

		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
	
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	} else {
		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	};

	sprintf(stmt, "delete from tb_binmatch where basedir = \"%s\" and filename = \"%s\";", ws->currentdir, ws->currentfile);
	run_statement(ws, ws->localdb[0], stmt);

	sprintf(stmt, "insert into tb_binmatch values ('%s', '%s', '%c', %zu, %d, 1);", ws->currentdir, ws->currentfile, cpe.part, strlen(cpe.vendor), rc);
	run_statement(ws, ws->localdb[0], stmt);

	// Match up to version (not edition and later)
	if ((swstrlen(cpe.update) == 0) && (swstrlen(cpe.edition) == 0) && (swstrlen(cpe.language) == 0))
		return;

	sprintf(stmt, "select count(*) from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"\" and cpeedition = \"\" and cpelanguage = \"\"", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version);

	rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);

	if (rc == 0) {
		sprintf(stmt, "insert into tb_cpe_%c_%zu (cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage) values (\"%c\", \"%s\", \"%s\", \"%s\", \"\", \"\", \"\");", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version);
		run_statement(ws, get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt);

		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"\" and cpeedition = \"\" and cpelanguage = \"\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version);
	
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	} else {
		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"\" and cpeedition = \"\" and cpelanguage = \"\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version);
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	};

	sprintf(stmt, "insert into tb_binmatch values ('%s', '%s', '%c', %zu, %d, 0);", ws->currentdir, ws->currentfile, cpe.part, strlen(cpe.vendor), rc);
	run_statement(ws, ws->localdb[0], stmt);

};


int file_already_processed(struct workstate * ws) {
 	int rc = 0;
	char stmt[1024];

	sprintf(stmt, "select count(*) from tb_binmatch where basedir = \"%s\" and filename = \"%s\";", ws->currentdir, ws->currentfile);

	rc = get_int_value(ws->localdb[0], stmt, ws);

	if (ws->rc != 0) {
		return 1;
	};

	return 0;
};

int display_cve_matches(void * cbobj, int argc, char **argv, char **azColName) {
	int i;
	int j;
	char * basedir = NULL;
	char * filename = NULL;
	char * cpe = NULL;
	char cpepart;
	char * cpevendorlength = NULL;
	int year = 0;
	int sequence = 0;
	int versionOnly = 0;

	j = 0;

	for (i = 0; i < argc; i++) {
		if (strcmp(azColName[i], "basedir") == 0) {
			basedir = argv[i];
			j++;
			continue;
		};
		if (strcmp(azColName[i], "filename") == 0) {
			filename = argv[i];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "year") == 0) {
			year = atoi(argv[i]);
			j++;
			continue;
		}
		if (strcmp(azColName[i], "sequence") == 0) {
			sequence = atoi(argv[i]);
			j++;
			continue;
		}
		if (strcmp(azColName[i], "cpe") == 0) {
			cpe = argv[i];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "cpepart") == 0) {
			cpepart = argv[i][0];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "cpevendorlength") == 0) {
			cpevendorlength = argv[i];
			j++;
			continue;
		}
		if (strcmp(azColName[i], "fullmatch") == 0) {
			versionOnly = atoi(argv[i]) == 0 ? 1 : 0;
			j++;
			continue;
		};
		fprintf(stderr, "Error: got column \"%s\" with no match.\n", azColName[i]);
	};

	if (j == 8) {
		char stmt[1024];
		int rc = 0;
		struct cpe_data cpedata;
		int length = atoi(cpevendorlength);
		struct workstate * ws = (struct workstate *) cbobj;

		sprintf(stmt, "select cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage from tb_cpe_%c_%d where cpeid = %s;", cpepart, length, cpe);
		rc = get_cpe_data(get_local_db(ws, cpepart, length), stmt, ws);
		cpedata = ws->cpebuffer;

		sprintf(stmt, "%s/%s", basedir, filename);

		show_potential_vulnerabilities(ws,  year,  sequence, stmt, cpedata, versionOnly);
	} else {
		fprintf(stderr, "Error, got %d entries back from query.\n", j);
	};

	return 0;
};


/**
 * Process the results of the file-matching query.
 *
 * The file matching query should provide a list of possible matches. These
 * matches in the master database contain information on how to extract the
 * version of the (matching) file. This function will then use this information
 * to obtain the version from the file and store it in the database.
 */
int get_version_and_store(void * cbobj, int argc, char **argv, char **azColName) {
	int i, rc;
	char * file_name_expression;
	char * file_content_expression;
	regex_t preg;
	regmatch_t pmatch[16];
	struct cpe_data cpe_data;
	struct workstate * ws = (struct workstate *) cbobj;
	int filetype = 0;

	ws->rc=0;	// Re-init state

	for(i=0;i<argc;i++) {
		// One or more hits
		if (strcmp(azColName[i], "filematch") == 0) {
			file_name_expression = argv[i];
			continue;
		};
		if (strcmp(azColName[i], "contentmatch") == 0) {
			file_content_expression = argv[i];
			continue;
		};
		if (strcmp(azColName[i], "cpepart") == 0) {
			cpe_data.part = argv[i][0];
			continue;
		};
		if (strcmp(azColName[i], "cpevendor") == 0) {
			strncpy(cpe_data.vendor, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeproduct") == 0) {
			strncpy(cpe_data.product, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeversion") == 0) {
			strncpy(cpe_data.version, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeupdate") == 0) {
			strncpy(cpe_data.update, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpeedition") == 0) {
			strncpy(cpe_data.edition, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "cpelanguage") == 0) {
			strncpy(cpe_data.language, argv[i], 64);
			continue;
		};
		if (strcmp(azColName[i], "filetype") == 0) {
			filetype = atoi(argv[i]);
			continue;
		};
	}

	// Verify if the filename matches file_name_expression
	rc = regcomp(&preg, file_name_expression, REG_EXTENDED);
	if (rc) {
		fprintf(stderr, "Failed to compile regular expression \"%s\"\n", file_name_expression);
		ws->rc = rc;
		return 0;
	};

	rc = regexec(&preg, ws->currentfile, 16, pmatch, 0);
	if (rc) {
		regfree(&preg);
		ws->rc = rc;
		return 0;
	};
	// Filename matches.

	// Free our memory allocations, and see if the content matches as well.
	regfree(&preg);
	rc = regcomp(&preg, file_content_expression, REG_EXTENDED);
	if (rc) {
		fprintf(stderr, "Failed to compile regular expression \"%s\"\n", file_content_expression);
		ws->rc = rc;
		return 0;
	};
	/**
	 * Here is where the various version extraction methods are supported.
	 * We currently still only support a single method (1, which is the
	 * "strings -n 3 <file>" command execution) but now we can see if we can
	 * support additional methods as well.
	 */
	if (filetype == 1) {
		char buffer[512];
		int ret;

		zero_string(buffer, 512);
		ret = strings_extract_version(ws, &preg, pmatch, &cpe_data);

		if (ret == 0) {
			add_to_database(ws, cpe_data);
			cpe_to_string(buffer, 512, cpe_data);
			fprintf(stdout, " - Found match for %s/%s:\t%s\n", ws->currentdir, ws->currentfile, buffer);
		};
	} else {
		fprintf(stderr, " ! %s/%s: The sqlite3 implementation currently doesn't support file type %d\n", ws->currentdir, ws->currentfile, filetype);
	};
	ws->rc = 0;
	return 0;
}

/**
 * Process the binary
 *
 * The binary file is provided through the workstate variable (currentfile and
 * currentdir properties).
 */
int dbimpl_process_binary(struct workstate * ws) {
	char buffer[1024];
	char * errmsg;

	int rc;

	zero_string(buffer, 1024);

	/*
	 * Query: find match in master database for the file. If match found,
	 * extract the version and store it in the local database.
	 */
	strcpy(buffer, "select v.filename as filename, v.filetype as filetype, v.filematch as filematch, v.contentmatch as contentmatch, c.cpepart as cpepart, c.cpevendor as cpevendor, c.cpeproduct as cpeproduct, c.cpeversion as cpeversion, c.cpeupdate as cpeupdate, c.cpeedition as cpeedition, c.cpelanguage as cpelanguage from tb_versionmatch v, tb_cpe c where v.cpe = c.cpeid and \"");
	strcat(buffer, ws->currentfile);
	strcat(buffer,"\" between filename and filename || \"z\";");

	rc = sqlite3_exec(ws->matchdb, buffer, get_version_and_store, ws, &errmsg);
	if (rc != SQLITE_OK) {
		fprintf(stderr, "SQL error: %s\n", errmsg);
		sqlite3_free(errmsg);
		return rc;
	};

	return 0;
};

void dbimpl_verify_installed_versus_cve(struct workstate * ws) {
	int rc = 0;
	char * errmsg;
	char stmt[1024];
	
	sprintf(stmt, "select b.basedir, b.filename, c.year, c.sequence, b.fullmatch, b.cpe, b.cpepart, b.cpevendorlength from tb_binmatch b, tb_cve c where b.cpe = c.cpe and b.cpepart = c.cpepart and b.cpevendorlength = c.cpevendorlength;");

	rc = sqlite3_exec(ws->localdb[0], stmt, display_cve_matches, ws, &errmsg);
	if (rc) {
		fprintf(stderr, "Failed to add information to database: %s\n", errmsg);
		sqlite3_free(errmsg);
	};
};

void dbimpl_initialize_workstate(struct workstate * ws) {
	int i = 0;

	ws->matchdb = NULL;
	ws->localdb = (sqlite3 **) calloc(sizeof(sqlite3 *), 64*3+1);
	for (i = 0; i <= 64*3; i++) {
		ws->localdb[i] = NULL;
	};
};

int check_cvecpe_in_db(struct workstate * ws, int year, int sequence, struct cpe_data cpe) {
	int rc = 0;
	char stmt[1024];

	sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);

	rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);

	if (rc == 0) {
		sprintf(stmt, "insert into tb_cpe_%c_%zu (cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage) values (\"%c\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\");", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language),
		run_statement(ws, get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt);
		sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);
	};

	sprintf(stmt, "select count(*) from tb_cve where year = %d and sequence = %d and cpepart = '%c' and cpevendorlength = %zu and cpe = %d;", year, sequence, cpe.part, strlen(cpe.vendor), rc);
	rc = get_int_value(ws->localdb[0], stmt, ws);
	return rc;
};

/**
 * Store the passed CVE entry in the database
 */
int dbimpl_store_cve_in_db(struct workstate * ws, char * cveId, char * cpeId) {
	int rc = 0;
	char stmt[1024];
	int year, sequence;
	struct cpe_data cpe;

	rc = cve_to_vars(&year, &sequence, cveId);
	if (rc == 1) {
		return 0;
	};
	string_to_cpe(&cpe, cpeId);

	ws->rc = 0;
	rc = check_cvecpe_in_db(ws, year, sequence, cpe);
	if ((rc != 0) && (ws->rc != 0)) {
		return 0;
	};

	sprintf(stmt, "select cpeid from tb_cpe_%c_%zu where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";",  cpe.part, strlen(cpe.vendor), cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
	rc = get_int_value(get_local_db(ws, cpe.part, strlen(cpe.vendor)), stmt, ws);

	sprintf(stmt, "insert into tb_cve values (%d, %d, '%c', %zu, %d);", year, sequence, cpe.part, strlen(cpe.vendor), rc);
	run_statement(ws, ws->localdb[0], stmt);

	return 1;
};


/**
 * Add versiongather entry into the database
 */
int dbimpl_add_versiongather(struct workstate * ws, struct versiongather_data vg, struct cpe_data cpe) {
	char stmt[1024];
	char stmt2[1024];
	int cpid = 0;

	sprintf(stmt, "select cpeid from tb_cpe where cpepart = \"%c\" and cpevendor = \"%s\" and cpeproduct = \"%s\" and cpeversion = \"%s\" and cpeupdate = \"%s\" and cpeedition = \"%s\" and cpelanguage = \"%s\";", cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);

	cpid = get_int_value(ws->matchdb, stmt, ws);

	if (cpid == 0) {
		zero_string(stmt2, 1024);
		sprintf(stmt2, "insert into tb_cpe (cpepart, cpevendor, cpeproduct, cpeversion, cpeupdate, cpeedition, cpelanguage) values (\"%c\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\");", cpe.part, cpe.vendor, cpe.product, cpe.version, cpe.update, cpe.edition, cpe.language);
		run_statement(ws, ws->matchdb, stmt2);

		cpid = get_int_value(ws->matchdb, stmt, ws);
	};

	if (cpid == 0) {
		fprintf(stderr, "Failed to store new data!\n");
		return 0;
	};

	zero_string(stmt2, 1024);
	sprintf(stmt2, "insert into tb_versionmatch values (\"%s\", %d, \"%s\", \"%s\", %d);", vg.filepart, vg.gathertype, vg.filematch, vg.versionexpression, cpid);
	run_statement(ws, ws->matchdb, stmt2);

	return 1;
};

/**
 * Initialize the databases
 */
int dbimpl_initialize_databases(struct workstate * ws) {
	char stmt[1024];
	int size = 1;

	// Setup of global
	sprintf(stmt, "PRAGMA foreign_keys=OFF; BEGIN TRANSACTION; DROP TABLE IF EXISTS tb_versionmatch; DROP TABLE IF EXISTS tb_cpe; CREATE TABLE tb_versionmatch (filename varchar(64), filetype smallint, filematch varchar(128), contentmatch varchar(256), cpe int); CREATE INDEX vmidx ON tb_versionmatch (filename); CREATE TABLE tb_cpe (cpeid integer primary key, cpepart char(1), cpevendor varchar(64), cpeproduct varchar(64), cpeversion varchar(64), cpeupdate varchar(64), cpeedition varchar(64), cpelanguage varchar(64)); COMMIT;");
	run_statement(ws, ws->matchdb, stmt);

	// Setup of locals
	zero_string(stmt, 1024);
	sprintf(stmt, "PRAGMA foreign_keys=OFF; BEGIN TRANSACTION; DROP TABLE IF EXISTS tb_binmatch; DROP TABLE IF EXISTS tb_cve; CREATE TABLE tb_binmatch ( basedir varchar(256), filename varchar(256), cpepart char(1), cpevendorlength int, cpe int, fullmatch int); CREATE TABLE tb_cve ( year smallint, sequence int, cpepart char(1), cpevendorlength int, cpe int); CREATE INDEX cveidx ON tb_cve (year, sequence); COMMIT;");
	run_statement(ws, ws->localdb[0], stmt);

	for (size = 1; size < 64; size++) {
		zero_string(stmt, 1024);
		sprintf(stmt, "DROP TABLE IF EXISTS tb_cpe_a_%d; CREATE TABLE tb_cpe_a_%d (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64));", size, size);
		run_statement(ws, get_local_db(ws, 'a', size), stmt);
	
		zero_string(stmt, 1024);
		sprintf(stmt, "DROP TABLE IF EXISTS tb_cpe_h_%d; CREATE TABLE tb_cpe_h_%d (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64));", size, size);
		run_statement(ws, get_local_db(ws, 'h', size), stmt);
		
		zero_string(stmt, 1024);
		sprintf(stmt, "DROP TABLE IF EXISTS tb_cpe_o_%d; CREATE TABLE tb_cpe_o_%d (cpeid integer primary key, cpepart char(1), cpevendor char(64), cpeproduct char(64), cpeversion char(64), cpeupdate char(64), cpeedition char(64), cpelanguage char(64));", size, size);
		run_statement(ws, get_local_db(ws, 'o', size), stmt);
	};

	return 1;
};

/**
 * Initialization for storing CVE data in the database.
 *
 * The initialization function is meant to allow implementations
 * to optimize their functions (like starting a transaction scope).
 */
void dbimpl_store_cve_in_db_init(struct workstate * ws) {
	// do nothing
};

/**
 * Exit method for finalizing the storage of CVE data in the database.
 *
 * The exit method is meant to allow implementations to optimize their functions
 * (like committing at the transaction scope).
 */
void dbimpl_store_cve_in_db_exit(struct workstate * ws) {
	// do nothing
};

/**
 * Initialize arguments structure with database-specific fields
 */
void dbimpl_initialize_arguments(struct arguments * arg) {
	// do nothing
};
